package foraging;

import com.cyberbotics.webots.controller.Servo;


public class Gripper extends foraging_stub.Gripper_stub {
	
	static public final byte CMD_PICK_UP = 0, CMD_RELEASE = 1;
    static public final byte ST_IDLE = 0, ST_CARRYING = 1, ST_PICKING = 2, ST_RELEASING = 3;
    
    static private final float POS_LIFT_UP = -1.4f, POS_LIFT_DOWN = 0f,
                               POS_GRIP_OPEN = 0.029f, POS_GRIP_CLOSED = 0.001f;
    
    private final Servo liftGrip;
    private final Servo leftGrip;
    private final Servo rightGrip;
    private final FSM fsm;
    private Byte cmd;
    
    
    public Gripper(Servo liftGrip, Servo leftGrip, Servo rightGrip) {
    	super();
        this.liftGrip = liftGrip;
        this.leftGrip = leftGrip;
        this.rightGrip = rightGrip;
        fsm = new FSM(IDLE, DOWN_PICK, CLOSE, UP_PICK,  // pick up
                      CARRY, DOWN_REL, OPEN, UP_REL);  // release
        cmd = null;
    }
    
    
    @Override
    public void init() {
        fsm.init();
    }
    
    
    @Override
    public void step() {
        cmd = _cmd.get();
        fsm.exec();
    }
    
    
    private boolean cmdPick() {
        return (cmd != null  &&  cmd == CMD_PICK_UP);
    }
    
    
    private boolean cmdRel() {
        return (cmd != null  &&  cmd == CMD_RELEASE);
    }
    
    
    private boolean isDown() {
        return (Math.abs(liftGrip.getPosition() - POS_LIFT_DOWN) < 0.01);
    }
    
    
    private boolean isUp() {
        return (Math.abs(liftGrip.getPosition() - POS_LIFT_UP) < 0.01);
    }
    
    
    private boolean isClosed() {
        return (Math.abs(leftGrip.getPosition() - POS_GRIP_CLOSED) < 0.002);
    }
    
    
    private boolean isOpen() {
        return (Math.abs(leftGrip.getPosition() - POS_GRIP_OPEN) < 0.002);
    }
    
    
    private boolean fail() {
        return leftGrip.getMotorForceFeedback() > -1f;
    }
    
    
    private final FSM.State IDLE = new FSM.State() {
        @Override
        protected void entry() {
            System.out.println("Gripper: idle");
            _status.send(ST_IDLE);
        }
        
        @Override
        protected void exec() {
            if (cmdPick()) {
                transition(DOWN_PICK);
            } else {
                _status.send(ST_IDLE);
            }
        }
    };
    
    
    private final FSM.State DOWN_PICK = new FSM.State() {
        @Override
        protected void entry() {
            System.out.println("Gripper: move down");
            liftGrip.setPosition(POS_LIFT_DOWN);
            _status.send(ST_PICKING);
        }
        
        @Override
        protected void exec() {
            if (cmdRel()) {
                transition(UP_REL);
            } else if (isDown()) {
                transition(CLOSE);
            } else {
                _status.send(ST_PICKING);
            }
        }
    };
    
    
    private final FSM.State CLOSE = new FSM.State() {
        private float avg;
        
        @Override
        protected void entry() {
            System.out.println("Gripper: close");
            leftGrip.setPosition(POS_GRIP_CLOSED);
            rightGrip.setPosition(POS_GRIP_CLOSED);
            avg = (float)leftGrip.getMotorForceFeedback();
            _status.send(ST_PICKING);
        }
        
        @Override
        protected void exec() {
            final float RATIO = 0.6f;
            float f = (float)leftGrip.getMotorForceFeedback();
            avg = RATIO*f + (1-RATIO)*avg;
            //System.out.println("Gripper: force = " + leftGrip.getMotorForceFeedback() + "  avg = " + avg);
            if (isClosed()  &&  fail()) {  // no object has been grabbed
                System.out.println("Gripper: no object grabbed!");
                transition(OPEN);
            } else if (cmdRel()) {
                transition(OPEN);
            } else if (Math.abs(f - avg) < 2  &&  f < -5) {  // an object has been properly grabbed
                transition(UP_PICK);
            } else {
                _status.send(ST_PICKING);
            }
        }
    };
    
    
    private final FSM.State UP_PICK = new FSM.State() {
        @Override
        protected void entry() {
            System.out.println("Gripper: lift up");
            liftGrip.setPosition(POS_LIFT_UP);
            _status.send(ST_PICKING);
        }
        
        @Override
        protected void exec() {
            //System.out.println("Gripper: force = " + leftGrip.getMotorForceFeedback());
            if (fail()) {  // the object has fallen
                System.out.println("Gripper: the object has fallen!");
                transition(OPEN);
            } else if (cmdRel()) {
                transition(DOWN_REL);
            } else if (isUp()) {
                transition(CARRY);
            } else {
                _status.send(ST_PICKING);
            }
        }
    };
    
    
    private final FSM.State CARRY = new FSM.State() {
        @Override
        protected void entry() {
            System.out.println("Gripper: carry");
            _status.send(ST_CARRYING);
        }
        
        @Override
        protected void exec() {
            //System.out.println("Gripper: force = " + leftGrip.getMotorForceFeedback());
            if (fail()) {  // the object has fallen
                System.out.println("Gripper: the object has fallen!");
                transition(OPEN);
            } else if (cmdRel()) {
                transition(DOWN_REL);
            } else {
                _status.send(ST_CARRYING);
            }
        }
    };
    
    
    private final FSM.State DOWN_REL = new FSM.State() {
        @Override
        protected void entry() {
            System.out.println("Gripper: move down");
            liftGrip.setPosition(POS_LIFT_DOWN);
            _status.send(ST_RELEASING);
        }
        
        @Override
        protected void exec() {
            //System.out.println("Gripper: force = " + leftGrip.getMotorForceFeedback());
            if (fail()) {  // the object has fallen
                System.out.println("Gripper: the object has fallen!");
                transition(OPEN);
            } else if (cmdPick()) {
                transition(UP_PICK);
            } else if (isDown()) {
                transition(OPEN);
            } else {
                _status.send(ST_RELEASING);
            }
        }
    };
    
    
    private final FSM.State OPEN = new FSM.State() {
        @Override
        protected void entry() {
            System.out.println("Gripper: open");
            leftGrip.setPosition(POS_GRIP_OPEN);
            rightGrip.setPosition(POS_GRIP_OPEN);
            _status.send(ST_RELEASING);
        }
        
        @Override
        protected void exec() {
            //System.out.println("Gripper: force = " + leftGrip.getMotorForceFeedback());
            if (cmdPick()) {
                transition(CLOSE);
            } else if (isOpen()) {
                transition(UP_REL);
            } else {
                _status.send(ST_RELEASING);
            }
        }
    };
    
    
    private final FSM.State UP_REL = new FSM.State() {
        @Override
        protected void entry() {
            System.out.println("Gripper: lift up");
            liftGrip.setPosition(POS_LIFT_UP);
            _status.send(ST_RELEASING);
        }
        
        @Override
        protected void exec() {
            if (cmdPick()) {
                transition(DOWN_PICK);
            } else if (isUp()) {
                transition(IDLE);
            } else {
                _status.send(ST_RELEASING);
            }
        }
    };
    
}
